/*
Copyright 2018 Toyota Research Institute North America


Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation and/or
 other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors 
may be used to endorse or promote products derived from this software without 
specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package fastsimj2;

import java.util.ArrayList;

import fastsimj2.components.*;

public class FASTSimJ2 {
	
	//Simulation constants
	private FSJSimConstants simConsts;
	
	
	//Fuel converter model
	private FuelConverterModel fcModel;
	//Electric motor model
	private ElectMotorModel mtModel;
	//Last Vehicle state
	private VehicleState lastVehState;
	
	//Last Simulation compact summary
	private TripCSummary lastSimSummary;
	public TripCSummary lastSimSummary() {return lastSimSummary;}
	//Recorded Vehicle States -- only has data after runF(...)
	private VehicleState[] vehStates;
	public VehicleState[] getLastSimVehStates() {return vehStates;}
	
	
	//Constructor
	public FASTSimJ2(FSJSimConstants simConstants) {
		simConsts = simConstants;
		
		fcModel = null;
		mtModel = null;
		lastVehState = null;
		lastSimSummary = new TripCSummary();
		
		vehStates = new VehicleState[0];
	}

	//Function to run "compact" simulation (only records/reports distance, fuel and battery use)
	// timeSec can be null (defaults to 1 sec intervals)
	// roadGrade can be null (defaults to zero)
	// additionalAuxPower can be null (defaults to zero)
	// initialRelSoC only affects BEVs, PHEVs and HEVs
	//    for B/PHEVs, initialRelSoC between 0 and 1 sets the initial SoC to relative fraction between minSoC and maxSoC
	//                 otherwise continues with last SoC from previous trip (if same vehicle model), otherwise sets SoC to maxSoC
	//    for HEVs, initialRelSoC >=0 attempts (iteratively) to match SoC at the beginning and end of the trip
	//              otherwise conducts 5 simulations with different initial SoC values, and "fits" equivalent fuel so that net battery use is zero
	public void runC(FSJVehicleModelParameters vehModel, float[] timeSec, float[] speedMPH, float[] roadGrade, 
			float[] additionalAuxPower, float initialRelSoC) {
		//Initialize power-train elements
		setupDriveModels(vehModel);
		//Reset detailed states record
		vehStates = new VehicleState[0];
		
		//Initialize vehicle state
		if (lastVehState==null) lastVehState = new VehicleState(simConsts, vehModel, fcModel, mtModel);
		else {
			if (!lastVehState.vehModel().name.equalsIgnoreCase(vehModel.name)) lastVehState = new VehicleState(simConsts, vehModel, fcModel, mtModel);
			else {
				lastVehState.resetAllButSoC(0f, 0f, 0f);
				if ((vehModel.vehPtType == VehicleDriveTrainType.bev)||(vehModel.vehPtType == VehicleDriveTrainType.phev)) {
					if ((initialRelSoC>=0)&&(initialRelSoC<=1)) lastVehState.setRelSoC(initialRelSoC);
				}
			}
		}		
		//HEV SoC matching
		if (vehModel.vehPtType == VehicleDriveTrainType.hev) {
			if (initialRelSoC < 0) {
				runHEV_FitSoC(timeSec, speedMPH, roadGrade, additionalAuxPower);
				return;
			}
			else {
				float iRelSoC = findBestHEVInitSoC(timeSec, speedMPH, roadGrade, additionalAuxPower);
				lastVehState.setRelSoC(iRelSoC);
			}
		}
		
		//Set initial state
		float ti = 0f;
		if (timeSec!=null) ti = timeSec[0];
		float mphi = speedMPH[0];
		float gradei = 0f;
		if (roadGrade!=null) gradei = roadGrade[0];
		
		lastVehState.resetAllButSoC(ti, mphi, gradei);
		lastSimSummary.reset();
		
		float speedMPHZTol = 0.001f;
		
		//Iterate
		for (int i=1; i<speedMPH.length; i++) {
			//Next step values
			float t = i;
			float deltaT = 1f;
			if (timeSec!=null) {
				t = timeSec[i];
				deltaT = t - timeSec[i-1];
			}
			float mph = speedMPH[i];
			float grade = 0f;
			if (roadGrade!=null) grade = roadGrade[i];
			float aaux = 0f;
			if (additionalAuxPower!=null) aaux = additionalAuxPower[i];
			
			//Update state
			lastVehState.updateState(t, mph, grade, aaux);
			
			//Update record of trip
			lastSimSummary.miles += lastVehState.distSLSMiles;
			lastSimSummary.seconds += deltaT;
			if (lastVehState.speedMPH < speedMPHZTol) lastSimSummary.secondsIdling += deltaT;
			lastSimSummary.fuelUse += lastVehState.fuelUseSLS;
			lastSimSummary.batteryUse += lastVehState.batteryUseSLS;
			lastSimSummary.maxSpeedSlipMPH = Math.max(lastSimSummary.maxSpeedSlipMPH, Math.abs(lastVehState.speedTraceSlipMPH));
			lastSimSummary.finalRelSoC = lastVehState.relSoC;
		}
	}
	//Function to run "full record" simulation (primary use for debugging and/or demo/presentations)
	// timeSec can be null (defaults to 1 sec intervals)
	// roadGrade can be null (defaults to zero)
	// additionalAuxPower can be null (defaults to zero)
	// initialRelSoC only affects BEVs, PHEVs and HEVs
	//    for B/PHEVs, initialRelSoC between 0 and 1 sets the initial SoC to relative fraction between minSoC and maxSoC
	//                 otherwise continues with last SoC from previous trip (if same vehicle model), otherwise sets SoC to maxSoC
	//    for HEVs, attempts (iteratively) to match SoC at the beginning and end of the trip
	public void runF(FSJVehicleModelParameters vehModel, float[] timeSec, float[] speedMPH, float[] roadGrade, 
			float[] additionalAuxPower, float initialRelSoC) {
		//Initialize power-train elements
		setupDriveModels(vehModel);
		
		//Initialize list for recording states
		ArrayList<VehicleState> lst = new ArrayList<VehicleState>();

		//Initialize vehicle state
		if (lastVehState==null) lastVehState = new VehicleState(simConsts, vehModel, fcModel, mtModel);
		else {
			if (!lastVehState.vehModel().name.equalsIgnoreCase(vehModel.name)) lastVehState = new VehicleState(simConsts, vehModel, fcModel, mtModel);
			else {
				lastVehState.resetAllButSoC(0f, 0f, 0f);
				if ((vehModel.vehPtType == VehicleDriveTrainType.bev)||(vehModel.vehPtType == VehicleDriveTrainType.phev)) {
					if ((initialRelSoC>=0)&&(initialRelSoC<=1)) lastVehState.setRelSoC(initialRelSoC);
				}
			}
		}		
		//HEV SoC matching
		if (vehModel.vehPtType == VehicleDriveTrainType.hev) {
			float iRelSoC = findBestHEVInitSoC(timeSec, speedMPH, roadGrade, additionalAuxPower);
			lastVehState.setRelSoC(iRelSoC);
		}
		
		
		//Set initial state
		float ti = 0f;
		if (timeSec!=null) ti = timeSec[0];
		float mphi = speedMPH[0];
		float gradei = 0f;
		if (roadGrade!=null) gradei = roadGrade[0];
		lastVehState.resetAllButSoC(ti, mphi, gradei);
		
		lst.add(new VehicleState(lastVehState));
		lastSimSummary.reset();
		
		float speedMPHZTol = 0.001f;
		
		//Iterate
		for (int i=1; i<speedMPH.length; i++) {
			//Next step values
			float t = i;
			float deltaT = 1f;
			if (timeSec!=null) {
				t = timeSec[i];
				deltaT = t - timeSec[i-1];
			}
			float mph = speedMPH[i];
			float grade = 0f;
			if (roadGrade!=null) grade = roadGrade[i];
			float aaux = 0f;
			if (additionalAuxPower!=null) aaux = additionalAuxPower[i];
			
			//Update state
			lastVehState.updateState(t, mph, grade, aaux);
			
			//Record details
			lst.add(new VehicleState(lastVehState));
			
			//Update record of trip
			lastSimSummary.miles += lastVehState.distSLSMiles;
			lastSimSummary.seconds += deltaT;
			if (lastVehState.speedMPH < speedMPHZTol) lastSimSummary.secondsIdling += deltaT;
			lastSimSummary.fuelUse += lastVehState.fuelUseSLS;
			lastSimSummary.batteryUse += lastVehState.batteryUseSLS;
			lastSimSummary.maxSpeedSlipMPH = Math.max(lastSimSummary.maxSpeedSlipMPH, Math.abs(lastVehState.speedTraceSlipMPH));
			lastSimSummary.finalRelSoC = lastVehState.relSoC;
		}
		
		//Copy to array
		vehStates = new VehicleState[lst.size()];
		for (int i=0; i<vehStates.length; i++) vehStates[i] = lst.get(i);
	}
	
	//Function to perform a run for HEV model with preset initial SoC
	private void runHEV_InitSoC(float[] timeSec, float[] speedMPH, float[] roadGrade, float[] additionalAuxPower, float initialRelSoC) {
		//Set initial state
		float ti = 0f;
		if (timeSec!=null) ti = timeSec[0];
		float mphi = speedMPH[0];
		float gradei = 0f;
		if (roadGrade!=null) gradei = roadGrade[0];
		
		lastVehState.resetAllButSoC(ti, mphi, gradei);
		lastVehState.setRelSoC(initialRelSoC);
		lastSimSummary.reset();
		
		float speedMPHZTol = 0.001f;
		
		//Iterate
		for (int i=1; i<speedMPH.length; i++) {
			//Next step values
			float t = i;
			float deltaT = 1f;
			if (timeSec!=null) {
				t = timeSec[i];
				deltaT = t - timeSec[i-1];
			}
			float mph = speedMPH[i];
			float grade = 0f;
			if (roadGrade!=null) grade = roadGrade[i];
			float aaux = 0f;
			if (additionalAuxPower!=null) aaux = additionalAuxPower[i];
			
			//Update state
			lastVehState.updateState(t, mph, grade, aaux);
			
			//Update record of trip
			lastSimSummary.miles += lastVehState.distSLSMiles;
			lastSimSummary.seconds += deltaT;
			if (lastVehState.speedMPH < speedMPHZTol) lastSimSummary.secondsIdling += deltaT;
			lastSimSummary.fuelUse += lastVehState.fuelUseSLS;
			lastSimSummary.batteryUse += lastVehState.batteryUseSLS;
			lastSimSummary.maxSpeedSlipMPH = Math.max(lastSimSummary.maxSpeedSlipMPH, Math.abs(lastVehState.speedTraceSlipMPH));
			lastSimSummary.finalRelSoC = lastVehState.relSoC;
		}
	}
	//Function to identify best initial SoC for HEV
	private float findBestHEVInitSoC(float[] timeSec, float[] speedMPH, float[] roadGrade, float[] additionalAuxPower) {
		int nMaxIter = 12;
		float isocZTol = 0.0001f;
		
		float isocLB = 0f;
		runHEV_InitSoC(timeSec, speedMPH, roadGrade, additionalAuxPower, isocLB);
		float dsocLB = lastVehState.relSoC - isocLB;
		if (dsocLB <= 0) return isocLB;
		
		float isocUB = 1f;
		runHEV_InitSoC(timeSec, speedMPH, roadGrade, additionalAuxPower, isocUB);
		float dsocUB = lastVehState.relSoC - isocUB;
		if (dsocUB >= 0) return isocUB;
		
		float isocTry = 0.5f*(isocUB-isocLB);
		runHEV_InitSoC(timeSec, speedMPH, roadGrade, additionalAuxPower, isocTry);
		float dsocTry = lastVehState.relSoC - isocTry;
		
		for (int i=1; i<nMaxIter; i++) {
			if (Math.abs(dsocTry)<=isocZTol) {
				return isocTry;
			}
			
			if (dsocTry > 0) {
				isocLB = isocTry;
				dsocLB = dsocTry;
			}
			else {
				isocUB = isocTry;
				dsocUB = dsocTry;
			}
			
			isocTry = 0.5f*(isocUB-isocLB);
			runHEV_InitSoC(timeSec, speedMPH, roadGrade, additionalAuxPower, isocTry);
			dsocTry = lastVehState.relSoC - isocTry;
		}		

		return isocTry;
	}	
	//Function to perform compact HEV run with fitting
	private void runHEV_FitSoC(float[] timeSec, float[] speedMPH, float[] roadGrade, float[] additionalAuxPower) {
		float d = 0;
		float zTol = 0.001f;		
		int n = 3;
		int nMax = 11;
		
		while (d < zTol) {
			n += 2;
			if (n>nMax) break;
			
			float sumX = 0f;
			float sumX2 = 0f;
			float sumY = 0f;
			float sumXY = 0f;
			float sumMiles = 0f;
			float maxSpeedSlip = 0f;
			float sumSecIdling = 0f;
			
			for (int i=0; i<n; i++) {
				float curISoC = (float)i/(float)(n-1);
				runHEV_InitSoC(timeSec, speedMPH, roadGrade, additionalAuxPower, curISoC);
				
				float x = curISoC - lastVehState.relSoC;
				float y = lastSimSummary.fuelUse;
				
				sumMiles += lastSimSummary.miles;
				sumSecIdling += lastSimSummary.secondsIdling;
				maxSpeedSlip = Math.max(maxSpeedSlip, lastSimSummary.maxSpeedSlipMPH);
				
				sumX += x;
				sumX2 += x*x;
				sumY += y;
				sumXY += x*y;
			}
			
			d = n*sumX2 - sumX*sumX;
			float d2 = sumX2*sumY - sumX*sumXY;
			
			lastSimSummary.miles = sumMiles/(float)n;
			lastSimSummary.secondsIdling = sumSecIdling/(float)n;
			lastSimSummary.fuelUse = d2/d;
			lastSimSummary.batteryUse = 0f;
			lastSimSummary.maxSpeedSlipMPH = maxSpeedSlip;
			lastSimSummary.finalRelSoC = 0f;
		}
	}	
	//Class for trip compact summary record
	public class TripCSummary {
		public float miles,fuelUse,batteryUse,maxSpeedSlipMPH,finalRelSoC,seconds,secondsIdling;
		public TripCSummary() {reset();}
		public void reset() {
			miles = 0;
			fuelUse = 0;
			batteryUse = 0;
			maxSpeedSlipMPH = 0;
			finalRelSoC = 0;
			seconds = 0;
			secondsIdling = 0;
		}
		public float mpg() {
			if (fuelUse<=0) return -1;
			return miles/fuelUse;
		}
		public float kwhpm() {
			if (miles<=0) return -1;
			return batteryUse/miles;
		}
		public float fracIdling() {
			if (seconds<=0) return -1;
			return secondsIdling/seconds;
		}
	}
	
	
	//Function to setup drive system models
	private void setupDriveModels(FSJVehicleModelParameters vehModel) {
		switch (vehModel.vehPtType) {
		case bev:
			fcModel = null;
			mtModel = new ElectMotorModel(vehModel);
			break;
		case cv:
			fcModel = new FuelConverterModel(vehModel);
			mtModel = null;
			break;
		case hev:
			fcModel = new FuelConverterModel(vehModel);
			mtModel = new ElectMotorModel(vehModel);
			break;
		case phev:
			fcModel = new FuelConverterModel(vehModel);
			mtModel = new ElectMotorModel(vehModel);
			break;
		}
	}	
	//Function to perform a run for HEV model with preset initial SoC
	public void runFHEV_InitSoCAbs(FSJVehicleModelParameters vehModel, 
			float[] timeSec, float[] speedMPH, float[] roadGrade, float[] additionalAuxPower, float initialSoCAbs) {
		
		if (vehModel.vehPtType != VehicleDriveTrainType.hev) return;
		
		float relSoC = (initialSoCAbs-vehModel.minSoc)/(vehModel.maxSoc-vehModel.minSoc);
		if (relSoC < 0) relSoC = 0;
		if (relSoC > 1) relSoC = 1;
		
		//Initialize power-train elements
		setupDriveModels(vehModel);
		//Initialize list for recording states
		ArrayList<VehicleState> lst = new ArrayList<VehicleState>();

		//Initialize vehicle state
		lastVehState = new VehicleState(simConsts, vehModel, fcModel, mtModel);
		lastVehState.setRelSoC(relSoC);
			
		//Set initial state
		float ti = 0f;
		if (timeSec!=null) ti = timeSec[0];
		float mphi = speedMPH[0];
		float gradei = 0f;
		if (roadGrade!=null) gradei = roadGrade[0];
		lastVehState.resetAllButSoC(ti, mphi, gradei);
		
		lst.add(new VehicleState(lastVehState));
		lastSimSummary.reset();
		
		float speedMPHZTol = 0.001f;
		
		//Iterate
		for (int i=1; i<speedMPH.length; i++) {
			//Next step values
			float t = i;
			float deltaT = 1f;
			if (timeSec!=null) {
				t = timeSec[i];
				deltaT = t - timeSec[i-1];
			}
			float mph = speedMPH[i];
			float grade = 0f;
			if (roadGrade!=null) grade = roadGrade[i];
			float aaux = 0f;
			if (additionalAuxPower!=null) aaux = additionalAuxPower[i];
			
			//Update state
			lastVehState.updateState(t, mph, grade, aaux);
			
			//Record details
			lst.add(new VehicleState(lastVehState));
			
			//Update record of trip
			lastSimSummary.miles += lastVehState.distSLSMiles;
			lastSimSummary.seconds += deltaT;
			if (lastVehState.speedMPH < speedMPHZTol) lastSimSummary.secondsIdling += deltaT;
			lastSimSummary.fuelUse += lastVehState.fuelUseSLS;
			lastSimSummary.batteryUse += lastVehState.batteryUseSLS;
			lastSimSummary.maxSpeedSlipMPH = Math.max(lastSimSummary.maxSpeedSlipMPH, Math.abs(lastVehState.speedTraceSlipMPH));
			lastSimSummary.finalRelSoC = lastVehState.relSoC;
		}
		
		//Copy to array
		vehStates = new VehicleState[lst.size()];
		for (int i=0; i<vehStates.length; i++) vehStates[i] = lst.get(i);
	}
}
