/*
Copyright 2018 Toyota Research Institute North America


Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation and/or
 other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors 
may be used to endorse or promote products derived from this software without 
specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package fastsimj2;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.SwingConstants;
import javax.swing.border.Border;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.filechooser.FileNameExtensionFilter;

@SuppressWarnings("serial")
public class FASTSimJ2GUI extends JFrame implements ActionListener {
	
	//Data Objects
	private FSJVehModelsTable vehTable;
	private FASTSimJ2Run.SpeedTraceFile speedTraceData;
	
	//Constructor
	public FASTSimJ2GUI() {
		//Create objects
		vehTable = null;
		speedTraceData = null;
		
		dlgCSVLoadFile = new JFileChooser();
		dlgCSVLoadFile.setDialogType(JFileChooser.OPEN_DIALOG);
		dlgCSVLoadFile.setFileSelectionMode(JFileChooser.FILES_ONLY);
	    FileNameExtensionFilter fileExtFilter = new FileNameExtensionFilter("CSV Files", "csv");
	    dlgCSVLoadFile.setFileFilter(fileExtFilter);
		
		
		//Create GUI objects
        setContentPane(createContentPane());       
        setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        setLocation(startPosOffset, startPosOffset);
        setSize(winWidth, winHeight);
        setTitle("Java Implementation of FASTSim");
        setResizable(false);
        setVisible(true);
	}
	
	//Function to run simulation
	private void runsSim(int chosenVehModelID) {
		if (vehTable==null) return;
		if (speedTraceData==null) return;
		
		FSJVehicleModelParameters vehModel = vehTable.getVehicleModel(chosenVehModelID);
		FASTSimJ2 fsj = new FASTSimJ2(new FSJSimConstants());	

		fsj.runC(vehModel, speedTraceData.time, speedTraceData.speedMPH, speedTraceData.grade, null, -1);		
		putResultToScreen(fsj.lastSimSummary());
	}
	private void putResultToScreen(FASTSimJ2.TripCSummary cSummary) {
		lblResults[0].setText(""+cSummary.miles);
		lblResults[1].setText(""+cSummary.fuelUse);
		lblResults[2].setText(""+cSummary.batteryUse);
		lblResults[3].setText(""+cSummary.maxSpeedSlipMPH);
		lblResults[4].setText(""+cSummary.finalRelSoC);
		lblResults[5].setText(""+cSummary.seconds);
		lblResults[6].setText(""+cSummary.secondsIdling);
		lblResults[7].setText(""+cSummary.mpg());
		lblResults[8].setText(""+cSummary.kwhpm());
		lblResults[9].setText(""+cSummary.fracIdling());
	}
	
	
	//Function to clear output labels
	private void clearOutputText() {
		for (int i=0; i<lblResults.length; i++) lblResults[i].setText("");
	}
	
	//Function for processing loading of vehicle models
	private void loadVehModels(String fname) {
		try {
			vehTable = new FSJVehModelsTable(fname);
		} catch (Exception e) {return;}
		
		String[] listTags = new String[vehTable.numVehicleModels()];
		for (int i=0; i<listTags.length; i++) listTags[i] = vehTable.getVehicleModel(i).name;
		lstVehModels.setListData(listTags);
		clearOutputText();
	}
	//Function for processing loading of trip trace
	private void loadtripTrace(String fname) {		
		try {
			speedTraceData = new FASTSimJ2Run.SpeedTraceFile(fname);
		} catch (Exception e) {
			speedTraceData = null;
			return;
		}
		
		lblTraceImage.repaint();
		clearOutputText();
		
		int curSelection = lstVehModels.getSelectedIndex();
		if (curSelection>=0) runsSim(curSelection);
	}
	//Function to process a selection change on the vehicle models list
	private void updateSelChange() {
		int curSelection = lstVehModels.getSelectedIndex();
		if (curSelection>=0) runsSim(curSelection);
	}

	//Responding to button clicks
	@Override
	public void actionPerformed(ActionEvent event) {
		Object source = event.getSource();
		
		if (source == btnVehModels) {
			int returnVal = dlgCSVLoadFile.showOpenDialog(null);
			if (returnVal == JFileChooser.APPROVE_OPTION) {
				String fileName = dlgCSVLoadFile.getSelectedFile().getAbsolutePath();
				loadVehModels(fileName);
			}			
		}
		if (source == btnTripTrace) {
			int returnVal = dlgCSVLoadFile.showOpenDialog(null);
			if (returnVal == JFileChooser.APPROVE_OPTION) {
				String fileName = dlgCSVLoadFile.getSelectedFile().getAbsolutePath();
				loadtripTrace(fileName);
			}			
		}
	}	
	
	
	//GUI size tuning values
	private static int winHeightOffset = 33;
	private static int winWidthOffset = 5;
	private static int winMargin = 10;
	
	private static int startPosOffset = 80;
	private static int winWidth = 800;
	private static int winHeight = 600;
	
	private static int winDrawHeight = winHeight - winHeightOffset;
	private static int winDrawWidth = winWidth - winWidthOffset;
	
	private static int numOutValues = 10;
	private static int staticTextHeight = 16;
	private static int buttonsHeight = staticTextHeight+4;
	private static int editBoxHeight = staticTextHeight+2;
	private static int buttonsWidth = 80;
	private static int lineSpace = buttonsHeight+4;
	private static int listWidth = 240;

	
	//GUI Objects
	private JButton btnVehModels, btnTripTrace;
	private TripTraceJLabel lblTraceImage;
	private JLabel[] lblResults;
	private JList<String> lstVehModels;
	private JFileChooser dlgCSVLoadFile;
	
	//Creation of GUI layout
	private JPanel createContentPane() {
		//Create
		JPanel totalGUI = new JPanel();
        totalGUI.setLayout(null);
        
        //Simple black border
        Border blackBorder = BorderFactory.createLineBorder(Color.BLACK);
        
        //Dependent size calculations
        int traceImgWidth = winDrawWidth - listWidth - 3*winMargin - winMargin/2;
        int traceImgHeight = winDrawHeight - lineSpace - (numOutValues/2)*lineSpace - 3*winMargin;
        int listHeight = winDrawHeight - lineSpace - 2*winMargin;
        int outStLabelsWidth = ((traceImgWidth-2*winMargin)/2) - buttonsWidth;
        
        //Top labels & Buttons
        int curX = winMargin;
        int curY = winMargin+(lineSpace-staticTextHeight)/2;
        
        JLabel lblSt1 = new JLabel("Vehicle Models");
        lblSt1.setLocation(curX, curY);
        lblSt1.setSize(listWidth-buttonsWidth, staticTextHeight);
        totalGUI.add(lblSt1);
        
        curX += listWidth-buttonsWidth;
        curY = winMargin+(lineSpace-buttonsHeight)/2;
        
        btnVehModels = new JButton("Load...");
        btnVehModels.setLocation(curX, curY);
        btnVehModels.setSize(buttonsWidth, buttonsHeight);
        btnVehModels.addActionListener(this);
        totalGUI.add(btnVehModels);
                
        
        curX = winDrawWidth-buttonsWidth-winMargin;
        btnTripTrace = new JButton("Load...");
        btnTripTrace.setLocation(curX, curY);
        btnTripTrace.setSize(buttonsWidth, buttonsHeight);
        btnTripTrace.addActionListener(this);
        totalGUI.add(btnTripTrace);
        
        curX = listWidth + 2*winMargin + winMargin/2;
        curY = winMargin+(lineSpace-staticTextHeight)/2;
        
        JLabel lblSt2 = new JLabel("Trip Trace");
        lblSt2.setLocation(curX, curY);
        lblSt2.setSize(traceImgWidth-buttonsWidth, staticTextHeight);
        totalGUI.add(lblSt2);
        
        //Image Label
        curY = winMargin + lineSpace;
        
        lblTraceImage = new TripTraceJLabel();
        lblTraceImage.setLocation(curX, curY);
        lblTraceImage.setSize(traceImgWidth, traceImgHeight);
        lblTraceImage.setBorder(blackBorder);
        totalGUI.add(lblTraceImage);
        
        //List of vehicle models
        curX = winMargin;
        
        lstVehModels = new JList<String>();
        lstVehModels.setLocation(0, 0);
        lstVehModels.setSize(listWidth, listHeight);
        lstVehModels.setBorder(blackBorder);
        lstVehModels.getSelectionModel().addListSelectionListener(new ListSelectionListener() {
        	public void valueChanged(ListSelectionEvent e) {
        		updateSelChange();
        	}
        });
    	JScrollPane scrollBar = new JScrollPane(lstVehModels);
    	scrollBar.setLocation(curX, curY);
    	scrollBar.setSize(listWidth, listHeight);  
    	totalGUI.add(scrollBar);
        
    	//Output value labels
    	String[] outCaps = outputCaptions();
    	lblResults = new JLabel[numOutValues];
    	JLabel[] lblSts = new JLabel[numOutValues];
        
    	for (int i=0; i<lblResults.length; i++) {
    		lblResults[i] = new JLabel();
    		lblSts[i] = new JLabel(outCaps[i]);
    		
    		lblResults[i].setSize(buttonsWidth, editBoxHeight);
    		lblResults[i].setBorder(blackBorder);
    		lblResults[i].setHorizontalAlignment(SwingConstants.RIGHT);
    		
    		lblSts[i].setSize(outStLabelsWidth, staticTextHeight);
    	}
    	
    	int halfNumLabels = numOutValues/2;
    	curX =  listWidth + 2*winMargin + winMargin/2;  	
    	curY = 2*winMargin + lineSpace + traceImgHeight;
    	
        for (int i=0; i<halfNumLabels; i++) {
        	lblSts[i].setLocation(curX, curY + (lineSpace-staticTextHeight)/2);
        	lblResults[i].setLocation(curX+outStLabelsWidth, curY + (lineSpace-editBoxHeight)/2);
        	
        	totalGUI.add(lblSts[i]);
        	totalGUI.add(lblResults[i]);
        	curY += lineSpace;
        }
        
    	curX += outStLabelsWidth + buttonsWidth + 2*winMargin;  	
    	curY = 2*winMargin + lineSpace + traceImgHeight;
    	
        for (int i=0; i<halfNumLabels; i++) {
        	lblSts[halfNumLabels+i].setLocation(curX, curY + (lineSpace-staticTextHeight)/2);
        	lblResults[halfNumLabels+i].setLocation(curX+outStLabelsWidth, curY + (lineSpace-editBoxHeight)/2);
        	
        	totalGUI.add(lblSts[halfNumLabels+i]);
        	totalGUI.add(lblResults[halfNumLabels+i]);
        	curY += lineSpace;
        }
    	
    	
        //Return
        totalGUI.setOpaque(true);
		return totalGUI;
	}	
	
	private static String[] outputCaptions() {
		String[] captions = {
				"Trip Length (miles)",
				"Fuel use (Gal)",
				"Battery use (kWh)",
				"Max. trace slip (mph)",
				"Ending Relative SoC",
				"Total duration (s)",
				"Time idling (s)",
				"Miles per gal",
				"kWh per mile",
				"Frac time idling"};
		return captions;
	}
	
	private class TripTraceJLabel extends JLabel {
		
		public TripTraceJLabel() {
			super();
		}
		
		private int allMargin = 5;
		private int axisTic = 5;
		private int legendLineTicLength = 30;
		private int nVAxesDiv = 10;
		
		private float fontSize = 10f; 
		
		private float minSpeed = 0f;
		private float maxSpeed = 100f;
		private float minGrade = -0.1f;
		private float maxGrade = 0.1f;

		private int leftMargin = 2*allMargin + (int)(2.5f*fontSize+0.5f) + axisTic;
		private int rightMargin = 2*allMargin + (int)(2*fontSize+0.5f) + axisTic;		
		private int topMargin = 2*allMargin + (int)(fontSize+0.5f);
		private int botMargin = allMargin + (int)(0.5*fontSize+0.5f);
		
		private int rightlegTextLength = (int)(6f*fontSize+0.5f);
		
		
		//Function to draw
		@Override
	    protected void paintComponent(Graphics g) {
			//Call base function
			super.paintComponent(g);
			
			//Width and height
			int w = getWidth();
			int h = getHeight();
			
			//White fill
			g.setColor(Color.WHITE);
			g.fillRect(0, 0, w, h);
			
			//Stop here if no trip trace has been loaded
			if (speedTraceData == null) return;

			
			//Draw
			g.setColor(Color.BLACK);
			g.setFont(g.getFont().deriveFont(fontSize));
			
			if (speedTraceData.grade != null) {
				g.setColor(Color.GRAY);
				int lastPlotX = xPos(0, speedTraceData.grade.length, w);
				int lastPlotY = vPosGrade(speedTraceData.grade[0], h);
				
				for (int i=1; i<speedTraceData.grade.length; i++) {
					int curPlotX = xPos(i, speedTraceData.grade.length, w);
					int curPlotY = vPosGrade(speedTraceData.grade[i], h);

					g.drawLine(lastPlotX, lastPlotY, curPlotX, curPlotY);

					lastPlotX = curPlotX;
					lastPlotY = curPlotY;
				}
			}
			if (speedTraceData.speedMPH != null) {
				g.setColor(Color.BLUE);
				int lastPlotX = xPos(0, speedTraceData.speedMPH.length, w);
				int lastPlotY = vPosSpeed(speedTraceData.speedMPH[0], h);
				
				for (int i=1; i<speedTraceData.speedMPH.length; i++) {
					int curPlotX = xPos(i, speedTraceData.speedMPH.length, w);
					int curPlotY = vPosSpeed(speedTraceData.speedMPH[i], h);

					g.drawLine(lastPlotX, lastPlotY, curPlotX, curPlotY);

					lastPlotX = curPlotX;
					lastPlotY = curPlotY;
				}
			}
			
			
			int llineYPos = (int)(allMargin+0.5f*fontSize+0.5f);
			g.setColor(Color.BLUE);
			g.drawLine(allMargin, llineYPos, allMargin+legendLineTicLength, llineYPos);
			g.setColor(Color.GRAY);
			g.drawLine(w-allMargin, llineYPos, w-allMargin-legendLineTicLength, llineYPos);
			
			g.setColor(Color.BLACK);
			g.drawString("Speed (mph)", 2*allMargin + legendLineTicLength, (int)(allMargin+fontSize+0.5f));
			g.drawString("Road Grade", w - 2*allMargin - legendLineTicLength - rightlegTextLength, (int)(allMargin+fontSize+0.5f));
			
			for (int i=0; i<=nVAxesDiv; i++) {
				float speedValue = minSpeed + (float)i*(maxSpeed-minSpeed)/(float)nVAxesDiv;
				float gradeValue = minGrade + (float)i*(maxGrade-minGrade)/(float)nVAxesDiv;
				int yPos = vPosSpeed(speedValue, h);
				g.drawLine(leftMargin, yPos, leftMargin-axisTic, yPos);
				g.drawLine(w-rightMargin, yPos, w-rightMargin+axisTic, yPos);
				
				yPos += (int)(0.5f*fontSize);
				g.drawString(""+speedValue,allMargin,yPos);
				g.drawString(""+gradeValue,w-rightMargin+axisTic+allMargin,yPos);
			}
			
			g.drawLine(leftMargin, topMargin, w-rightMargin, topMargin);
			g.drawLine(w-rightMargin, topMargin, w-rightMargin, h-botMargin);
			g.drawLine(w-rightMargin, h-botMargin, leftMargin, h-botMargin);
			g.drawLine(leftMargin, h-botMargin, leftMargin, topMargin);
		}	
		
		private int vPosSpeed(float speedMPH, int h) {
			float yRel = (speedMPH-minSpeed)/(maxSpeed-minSpeed);
			return h - botMargin - (int)((h-topMargin-botMargin)*yRel + 0.5f);
		}
		private int vPosGrade(float grade, int h) {
			float yRel = (grade-minGrade)/(maxGrade-minGrade);
			return h - botMargin - (int)((h-topMargin-botMargin)*yRel + 0.5f);
		}
		private int xPos(int arrayID, int arrayLength, int w) {
			float xRel = (float)arrayID/(float)(arrayLength-1);
			return leftMargin + (int)(xRel*(w-leftMargin-rightMargin)+0.5);
		}
	}
	
}
