/*
Copyright 2018 Toyota Research Institute North America


Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation and/or
 other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors 
may be used to endorse or promote products derived from this software without 
specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package fastsimj2;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;

public class FASTSimJ2Run {

	public static void main(String[] args) {
		
		switch (args.length) {
		case 4:
			//Single speed trace
			runSingle(args);
			break;
		case 5:
			//Multiple speed traces (all assumed on the same day)
			runMultiple(args);
			break;
		default:
			new FASTSimJ2GUI();
		}
	}
	
	private static void runSingle(String[] args) {
		String vehicleModelsFile = args[0];
		int chosenVehicleID = Integer.parseInt(args[1]);
		String outputFile = args[2];
		String speedTraceFile = args[3];
		
		//Initialize
		FSJVehModelsTable vehTable = new FSJVehModelsTable(vehicleModelsFile);
		FSJVehicleModelParameters vehModel = vehTable.getVehicleModel(chosenVehicleID);
		FASTSimJ2 fsj = new FASTSimJ2(new FSJSimConstants());		
		
		//Relay Input
		System.out.println("Vehicles Table: "+vehicleModelsFile);
		System.out.println("Chosen Vehicle ID: "+chosenVehicleID+" ("+vehModel.name+")");
		
		//Read Speed trace file
		System.out.println("Speed Trace: "+speedTraceFile);
		SpeedTraceFile spdTrace = new SpeedTraceFile(speedTraceFile);
		
		//Solve
		fsj.runC(vehModel, spdTrace.time, spdTrace.speedMPH, spdTrace.grade, null, -1);
		
		//Write output file
		System.out.println("Writing output file: "+outputFile);
		try {
			FileWriter fWriter = new FileWriter(outputFile);
			String lsep = System.getProperty("line.separator");
			
			fWriter.append(tripSummaryHeader());
			fWriter.append(lsep);
			
			fWriter.append(tripSummaryString(fsj.lastSimSummary()));
			fWriter.append(lsep);
			
			fWriter.flush();
			fWriter.close();
		} catch (IOException e) {}
		
		//Indicate finishing
		System.out.println("Done!");
	}
	private static void runMultiple(String[] args) {
		String vehicleModelsFile = args[0];
		int chosenVehicleID = Integer.parseInt(args[1]);
		String outputFile = args[2];
		String speedTracesFile = args[3];
		int chargignModelID = Integer.parseInt(args[4]);		
		ChargingModel chgModel = ChargingModel.values()[chargignModelID];
		
		//Initialize
		FSJVehModelsTable vehTable = new FSJVehModelsTable(vehicleModelsFile);
		FSJVehicleModelParameters vehModel = vehTable.getVehicleModel(chosenVehicleID);
		FASTSimJ2 fsj = new FASTSimJ2(new FSJSimConstants());		
		
		//Relay Input
		System.out.println("Vehicles Table: "+vehicleModelsFile);
		System.out.println("Chosen Vehicle ID: "+chosenVehicleID+" ("+vehModel.name+")");

		//Main loop
		SpeedTracesBatch spdTracesInfo = new SpeedTracesBatch(speedTracesFile);
		
		System.out.println("Output file: "+outputFile);
		try {
			FileWriter fWriter = new FileWriter(outputFile);
			String lsep = System.getProperty("line.separator");
			
			fWriter.append(tripSummaryHeader());
			fWriter.append(lsep);
			
			int lastDayID = -1;
			
			for (int i=0; i<spdTracesInfo.speedTracesInfo.length; i++) {
				String curSpeedTraceFile = spdTracesInfo.speedTracesInfo[i].speedTraceFile;
				System.out.println("Speed Trace: "+curSpeedTraceFile);
				SpeedTraceFile spdTrace = new SpeedTraceFile(curSpeedTraceFile);
		
				switch (vehModel.vehPtType) {
				case bev:
				case phev:
					float relSoC = 0f;
					int curDayID = spdTracesInfo.speedTracesInfo[i].dayID;
					
					switch (chgModel) {
					case never:
						relSoC = 0f;
						break;
					case opp4hr:
						if (curDayID > lastDayID) {
							relSoC = 1f;
							lastDayID = curDayID;
						}
						else {
							float hrThreshold = 4f;
							if (spdTracesInfo.speedTracesInfo[i].hoursSinceLastTrip >= hrThreshold) {
								relSoC = 1f;
							}
							else {
								relSoC = -1;
							}
						}
						break;
					case overnight:
						if (curDayID > lastDayID) {
							relSoC = 1f;
							lastDayID = curDayID;
						}
						else {
							relSoC = -1;
						}
						break;
					case startFull:
						relSoC = 1f;
						break;
					}
					fsj.runC(vehModel, spdTrace.time, spdTrace.speedMPH, spdTrace.grade, null, relSoC);	
					if (i==0) fsj.runC(vehModel, spdTrace.time, spdTrace.speedMPH, spdTrace.grade, null, relSoC);
					break;
				default:	//CV & HEV
					fsj.runC(vehModel, spdTrace.time, spdTrace.speedMPH, spdTrace.grade, null, -1);
				}
			
				fWriter.append(tripSummaryString(fsj.lastSimSummary()));
				fWriter.append(lsep);
				fWriter.flush();
			}
			
			fWriter.close();
		} catch (IOException e) {}
		
		//Indicate finishing
		System.out.println("Done!");
	}
	
	
	public static String tripSummaryString(FASTSimJ2.TripCSummary cSummary) {
		String strOut = ""+cSummary.miles;
		strOut = strOut+","+cSummary.fuelUse;
		strOut = strOut+","+cSummary.batteryUse;
		strOut = strOut+","+cSummary.maxSpeedSlipMPH;
		strOut = strOut+","+cSummary.finalRelSoC;
		strOut = strOut+","+cSummary.seconds;
		strOut = strOut+","+cSummary.secondsIdling;
		strOut = strOut+","+cSummary.mpg();
		strOut = strOut+","+cSummary.kwhpm();
		strOut = strOut+","+cSummary.fracIdling();
		return strOut;
	}
	public static String tripSummaryHeader() {
		String strOut = "miles,fuelUse,batteryUse,maxSpeedSlipMPH,finalRelSoC,seconds,secondsIdling,mpg,kwhpm,fracIdling";
		return strOut;
	}
	
	public static class SpeedTraceFile {
		public float[] time,speedMPH,grade;
		
		public SpeedTraceFile(String fname) {
			ArrayList<Float> lstTime = new ArrayList<Float>();
			ArrayList<Float> lstSpeed = new ArrayList<Float>();
			ArrayList<Float> lstGrade = new ArrayList<Float>();		
			
			try {
				BufferedReader readingBuffer=new BufferedReader(new FileReader(fname));
				String readLine = readingBuffer.readLine();
				String[] strSplit = readLine.split(",");
				int numColumns = strSplit.length;
			
				switch (numColumns) {
				case 3:
					while ((readLine = readingBuffer.readLine())!=null) {
						strSplit = readLine.split(",");
						lstTime.add(Float.parseFloat(strSplit[0]));
						lstSpeed.add(Float.parseFloat(strSplit[1]));
						lstGrade.add(Float.parseFloat(strSplit[2]));
					}
					break;
				case 2:
					while ((readLine = readingBuffer.readLine())!=null) {
						strSplit = readLine.split(",");
						lstSpeed.add(Float.parseFloat(strSplit[0]));
						lstGrade.add(Float.parseFloat(strSplit[1]));
					}
					break;
				case 1:
					while ((readLine = readingBuffer.readLine())!=null) {
						lstSpeed.add(Float.parseFloat(readLine));
					}
				}
			
				readingBuffer.close();
			} catch (IOException e) {}
			
			speedMPH = new float[lstSpeed.size()];
			for (int i=0; i<speedMPH.length; i++) speedMPH[i] = lstSpeed.get(i);			
			time = null;
			if (lstTime.size()>0) {
				time = new float[lstTime.size()];
				for (int i=0; i<time.length; i++) time[i] = lstTime.get(i);
			}		
			grade = null;
			if (lstGrade.size()>0) {
				grade = new float[lstGrade.size()];
				for (int i=0; i<grade.length; i++) grade[i] = lstGrade.get(i);
			}
		}
	}

	public static class SpeedTraceInBatch {
		public int dayID;
		public float hoursSinceLastTrip;
		public String speedTraceFile;
		
		public SpeedTraceInBatch(String readLine) {
			String[] strSplit = readLine.split(",");
			dayID = Integer.parseInt(strSplit[0]);
			hoursSinceLastTrip = Float.parseFloat(strSplit[1]);
			speedTraceFile = strSplit[2];
		}
	}
	public static class SpeedTracesBatch {
		public SpeedTraceInBatch[] speedTracesInfo;
		public SpeedTracesBatch(String fname) {
			ArrayList<SpeedTraceInBatch> lst = new ArrayList<SpeedTraceInBatch>();
			
			try {
				BufferedReader readingBuffer=new BufferedReader(new FileReader(fname));
				String readLine = readingBuffer.readLine();
				
				while ((readLine = readingBuffer.readLine())!=null) {
					lst.add(new SpeedTraceInBatch(readLine));
				}
				
				readingBuffer.close();
			} catch (IOException e) {}
			
			speedTracesInfo = new SpeedTraceInBatch[lst.size()];
			for (int i=0; i<speedTracesInfo.length; i++) speedTracesInfo[i] = lst.get(i);
		}
	}
	public enum ChargingModel {
		startFull("Start full"),
		opp4hr("Opportune 4hr"),
		overnight("Overnight Only"),
		never("Never Charge"),
		;
		public String tag;
		private ChargingModel(String c) {
			tag = c;
		}
	}

}
