/*
Copyright 2018 Toyota Research Institute North America


Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation and/or
 other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors 
may be used to endorse or promote products derived from this software without 
specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package fastsimj2;

import fastsimj2.components.*;

public class FSJVehicleModelParameters {
	//Tag name for vehicle model
	public String name;
	//Type of drive (CV, HEV, PHEV, BEV)
	public VehicleDriveTrainType vehPtType;
	//Air drag coefficient
	public float dragCoef;
	//Front Projected Area
	public float frontalAreaM2;
	//Base weight of the vehicle
	public float gliderKg;
	//Center of gravity location in meters
	public float vehCgM;
	//Drive axle weight fraction
	public float driveAxleWeightFrac;
	//Wheel base in meters
	public float wheelBaseM;
	//Cargo weight
	public float cargoKg;
	//Maximum power draw (in kW) from fuel storage
	public float maxFuelStorKw;
	//Transition time to reach peak power from fuel storage
	public float fuelStorSecsToPeakPwr;
	//Fuel storage in kWh
	public float fuelStorKwh;
	//Fuel storage system kWh per kg
	public float fuelStorKwhPerKg;
	//Maximum fuel converter (i.e. internal combustion engine) power in kW
	public float maxFuelConvKw;
	//Type of fuel converter (Spark Ignition, Atkinson, Diesel)
	public FuelConverterEffType fcEffType;
	//Time for fuel converter to reach its maximum power
	public float fuelConvSecsToPeakPwr;
	//Unit weight (per kW) of fuel converter
	public float fuelConvKwPerKg;
	//Maximum power (in kW) of electric motor
	public float maxMotorKw;
	//Peak efficiency of electric motor
	public float motorPeakEff;
	//Time for electric motor to reach maximum power
	public float motorSecsToPeakPwr;
	//Unit weight (per kW) of electric motor
	public float mcPeKgPerKw;
	//Base weight of electric motor
	public float mcPeBaseKg;
	//Maximum battery power in kW
	public float maxEssKw;
	//Override flag for battery KW limit
	public boolean useMotorKwInsteadOfESS;
	//Maximum battery energy storage in kWh
	public float maxEssKwh;
	//Unit weight (per kW) of battery
	public float essKgPerKwh;
	//Base weight of battery
	public float essBaseKg;
	//"Round-trip" (i.e. charge and discharge) efficiency of battery 
	public float essRountTripEff;
	//Coefficient A for battery life
	public float essLifeCoefA;
	//Coefficient A for battery life
	public float essLifeCoefB;
	//Wheel inertia in kg.m^2
	public float wheelInteriaKgM2;
	//Number of wheels
	public float numWheels;
	//Wheel rolling resistance coefficient
	public float wheelRrCoef;
	//Wheel radius in meters
	public float wheelRadiusM;
	//Wheel coefficient of friction
	public float wheelCoefOfFric;
	//Energy management control strategy (kinetic energy, engine efficiency)
	public EnergyMgmtControlStrategy controlStrategy;
	//Minimum state of charge (as a fraction of maxEssKwh) -- this corresponds to "empty battery" in some other contexts
	public float minSoc;
	//Maximum state of charge (as a fraction of maxEssKwh) -- this corresponds to "full battery" in some other contexts
	public float maxSoc;
	//Tunable parameter for engine start/stop
	public float minSocMph;
	//Tunable parameter for engine start/stop
	public float maxSocMph;
	//Tunable parameter for percentage power assist
	public float assistKwPerPercOff;
	//Tunable parameter for charge off
	public float chgKwPerPercOff;
	//Tunable parameter for usable SoC
	public float percUsableSocPassBuf;
	//Tunable parameter for Fuel Converter power
	public float percFcPwrDoneElec;
	//Tunable parameter for high acceleration
	public float percHighAccBuf;
	//Speed for forcing Fuel Converter on
	public float mphFcOn;
	//Power demand for forcing fuel converter on
	public float kwDemandFcOn;
	//Alternator efficiency
	public float altEff;
	//Charging efficiency
	public float chgEff;
	//Auxiliary power demand in the vehicle
	public float auxKw;
	//Transmission system weight in kg
	public float transKg;
	//Transmission system efficiency
	public float transEff;
	//"acceptable error" of battery to fuel (only affects HEVs, where the simulation attempts to match the battery SoC at the beginning and end of each trip)
	public float essToFuelOkError;
	//Maximum regeneration (in regenerative braking?)
	public float maxRegen;
	
	//Constructor
	public FSJVehicleModelParameters(String readLine) {parseCSVLine(readLine);}
	public FSJVehicleModelParameters(FSJVehicleModelParameters other) {parseCSVLine(other.toString());}
	
	//Function to parse values from CSV line
	public void parseCSVLine(String readLine) {
		String[] strSplit = readLine.split(",");

		name = strSplit[0];
		vehPtType = VehicleDriveTrainType.values()[Integer.parseInt(strSplit[1])-1];

		dragCoef = Float.parseFloat(strSplit[2]);
		frontalAreaM2 = Float.parseFloat(strSplit[3]);
		gliderKg = Float.parseFloat(strSplit[4]);
		
		vehCgM = Float.parseFloat(strSplit[5]);
		driveAxleWeightFrac = Float.parseFloat(strSplit[6]);
		wheelBaseM = Float.parseFloat(strSplit[7]);
		cargoKg = Float.parseFloat(strSplit[8]);
		
		maxFuelStorKw = Float.parseFloat(strSplit[9]);		
		fuelStorSecsToPeakPwr = Float.parseFloat(strSplit[10]);
		fuelStorKwh = Float.parseFloat(strSplit[11]);
		fuelStorKwhPerKg = Float.parseFloat(strSplit[12]);
		
		maxFuelConvKw = Float.parseFloat(strSplit[13]);
		fcEffType = FuelConverterEffType.values()[Integer.parseInt(strSplit[14])-1];
		
		fuelConvSecsToPeakPwr = Float.parseFloat(strSplit[15]);
		fuelConvKwPerKg = Float.parseFloat(strSplit[16]);
		
		maxMotorKw = Float.parseFloat(strSplit[17]);
		motorPeakEff = Float.parseFloat(strSplit[18]);
		motorSecsToPeakPwr = Float.parseFloat(strSplit[19]);
		
		mcPeKgPerKw = Float.parseFloat(strSplit[20]);
		mcPeBaseKg = Float.parseFloat(strSplit[21]);
		
		maxEssKw = Float.parseFloat(strSplit[22]);
		
		useMotorKwInsteadOfESS = false;
		if (Integer.parseInt(strSplit[23])!=0) useMotorKwInsteadOfESS = true;
		
		maxEssKwh = Float.parseFloat(strSplit[24]);
		essKgPerKwh = Float.parseFloat(strSplit[25]);		
		essBaseKg = Float.parseFloat(strSplit[26]);
		essRountTripEff = Float.parseFloat(strSplit[27]);
		essLifeCoefA = Float.parseFloat(strSplit[28]);
		essLifeCoefB = Float.parseFloat(strSplit[29]);
		
		wheelInteriaKgM2 = Float.parseFloat(strSplit[30]);		
		numWheels = Float.parseFloat(strSplit[31]);
		wheelRrCoef = Float.parseFloat(strSplit[32]);
		wheelRadiusM = Float.parseFloat(strSplit[33]);
		wheelCoefOfFric = Float.parseFloat(strSplit[34]);
		
		controlStrategy = EnergyMgmtControlStrategy.values()[Integer.parseInt(strSplit[35])-1];
		
		minSoc = Float.parseFloat(strSplit[36]);
		maxSoc = Float.parseFloat(strSplit[37]);
		minSocMph = Float.parseFloat(strSplit[38]);
		maxSocMph = Float.parseFloat(strSplit[39]);
		assistKwPerPercOff = Float.parseFloat(strSplit[40]);
		
		chgKwPerPercOff = Float.parseFloat(strSplit[41]);
		percUsableSocPassBuf = Float.parseFloat(strSplit[42]);
		percFcPwrDoneElec = Float.parseFloat(strSplit[43]);
		percHighAccBuf = Float.parseFloat(strSplit[44]);
		mphFcOn = Float.parseFloat(strSplit[45]);		
		kwDemandFcOn = Float.parseFloat(strSplit[46]);
		
		altEff = Float.parseFloat(strSplit[47]);
		chgEff = Float.parseFloat(strSplit[48]);
		auxKw = Float.parseFloat(strSplit[49]);
		transKg = Float.parseFloat(strSplit[50]);		
		transEff = Float.parseFloat(strSplit[51]);
		essToFuelOkError = Float.parseFloat(strSplit[52]);
		maxRegen = Float.parseFloat(strSplit[53]);
	}

	//Function to form Equivalent String
	@Override public String toString() {
		int useMotorKwInsteadOfESSInt = 0;
		if (useMotorKwInsteadOfESS) useMotorKwInsteadOfESSInt = 1;
		return ""
				+ name + ","
				+ (vehPtType.ordinal()+1) + ","
				
				+ dragCoef + ","
				+ frontalAreaM2 + ","
				+ gliderKg + ","
				
				+ vehCgM + ","
				+ driveAxleWeightFrac + ","
				+ wheelBaseM + ","
				+ cargoKg + ","
				
				+ maxFuelStorKw + ","
				+ fuelStorSecsToPeakPwr + ","
				+ fuelStorKwh + ","
				+ fuelStorKwhPerKg + ","
				
				+ maxFuelConvKw + ","
				+ (fcEffType.ordinal()+1) + ","
				
				+ fuelConvSecsToPeakPwr + ","
				+ fuelConvKwPerKg + ","
				
				+ maxMotorKw + ","
				+ motorPeakEff + ","
				+ motorSecsToPeakPwr + ","
				
				+ mcPeKgPerKw + ","
				+ mcPeBaseKg + ","
				
				+ maxEssKw + ","
				+ useMotorKwInsteadOfESSInt + ","
				+ maxEssKwh + ","
				+ essKgPerKwh + ","
				+ essBaseKg + ","
				+ essRountTripEff + ","
				+ essLifeCoefA + ","
				+ essLifeCoefB + ","
				
				+ wheelInteriaKgM2 + ","
				+ numWheels + ","
				+ wheelRrCoef + ","
				+ wheelRadiusM + ","
				+ wheelCoefOfFric + ","
				
				+ (controlStrategy.ordinal()+1) + ","
				
				+ minSoc + ","
				+ maxSoc + ","
				+ minSocMph + ","
				+ maxSocMph + ","
				+ assistKwPerPercOff + ","
				
				+ chgKwPerPercOff + ","
				+ percUsableSocPassBuf + ","
				+ percFcPwrDoneElec + ","
				+ percHighAccBuf + ","
				+ mphFcOn + ","
				+ kwDemandFcOn + ","
				
				+ altEff + ","
				+ chgEff + ","
				+ auxKw + ","
				+ transKg + ","
				+ transEff + ","
				+ essToFuelOkError + ","
				+ maxRegen
				;
	}
}
