/*
Copyright 2018 Toyota Research Institute North America


Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation and/or
 other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors 
may be used to endorse or promote products derived from this software without 
specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package fastsimj2.components;

import fastsimj2.*;

public class ElectMotorModel {
	private float maxOutPowerKW, maxKWIn, minMotorEff;
	private float[] aCoeff, bCoeff;
	
	public float minMotorEff() {return minMotorEff;}
	
	public float inputPowerKW(float outPowerKW) {
		float x = outPowerKW;
		if (x < 0) x = 0;
		if (x > maxOutPowerKW) x = maxOutPowerKW;
		
		float inPow = 0f;
		float xp = 1f;
		
		for (int i=0; i<aCoeff.length; i++) {
			inPow += xp * aCoeff[aCoeff.length-1-i];
			xp *= x;
		}
		
		return inPow;
	}
	public float outputPowerKW(float inPowerKW) {
		float x = inPowerKW;
		if (x < 0) x = 0;
		if (x > maxKWIn) x = maxKWIn;
		
		float outPow = 0f;
		float xp = 1f;
		
		for (int i=0; i<bCoeff.length; i++) {
			outPow += xp * bCoeff[bCoeff.length-1-i];
			xp *= x;
		}
		
		return outPow;
	}
	
	private static int polyOrder() {return 4;}
	private static float[] fitPtsFracOutKW() {
		float[] fitsPts = {0.000f,
				0.001f,0.002f,0.003f,0.004f,0.005f,
				0.006f,0.007f,0.008f,0.01f,
				0.02f,0.03f,0.04f,0.05f,0.1f,0.15f,
				0.2f,0.25f,0.3f,0.35f,0.4f,0.45f,
				0.5f,0.55f,0.6f,0.65f,0.7f,0.75f,
				0.8f,0.85f,0.9f,0.95f,1.0f};
		return fitsPts;
	}
	private static float point2loadFrac(float maxKW) {
		float lineFitAdd = -0.06f*(float)Math.log(maxKW)+0.24f;
		return 0.09f + Math.max(lineFitAdd, 0);
	}

	
	public ElectMotorModel(FSJVehicleModelParameters vehModel) {
		maxOutPowerKW = vehModel.maxMotorKw;
		
		float[] x4 = {0.0f, point2loadFrac(maxOutPowerKW), 0.65f, 1.0f};
		float[] y4 = {0.3f, 0.98f*(vehModel.motorPeakEff-0.01f),  (vehModel.motorPeakEff-0.01f), 0.98f*(vehModel.motorPeakEff-0.01f)};
		
		minMotorEff = y4[0];
		
		maxKWIn = maxOutPowerKW/y4[y4.length-1];
		
		float[] powFracs = fitPtsFracOutKW();
		float[] x = new float[powFracs.length];
		float[] y = new float[powFracs.length];
		
		prepCoeff(x,y,x4,y4,powFracs);
		aCoeff = new float[polyOrder()+1];
		bCoeff = new float[polyOrder()+1];
		calcCoeff(x,y,aCoeff);
		calcCoeff(y,x,bCoeff);
	}
	private void prepCoeff(float[] x, float[] y, float[] x4, float[] y4, float[] powFracts) {
		for (int i=0; i < powFracts.length; i++) {
			x[i] = powFracts[i]*maxOutPowerKW;
			y[i] = x[i]/linInterpolate(x4,y4,powFracts[i]);
		}
	}
	private static float linInterpolate(float[] xs, float[] ys, float x) {
		if (x <= xs[0]) return ys[0];
		if (x >= xs[xs.length-1]) return ys[xs.length-1];
		
		int intervalID = 0;
		while (xs[intervalID+1] < x) intervalID++;
		
		float x1 = xs[intervalID];
		float x2 = xs[intervalID+1];
		float z = (x-x1)/(x2-x1);
		
		return ys[intervalID] + z*(ys[intervalID+1]-ys[intervalID]);
	}
	private static void calcCoeff(float[] x, float[] y, float[] coeff) {
		int nDim = polyOrder() + 1;
		if (nDim > x.length) nDim = x.length;
		
		float[] xps = new float[2*polyOrder() + 1];
		double[][] mat = new double[nDim][nDim];
		
		for (int k=0; k<x.length; k++) {	//Loop on points
			calcPows(xps, x[k]);	//Powers of x
			
			//Matrix coefficients
			for (int i=0; i<nDim; i++) {
				mat[i][i] += xps[2*i];
				
				for (int j=i+1; j<nDim; j++) {
					mat[i][j] += xps[i+j];
					mat[j][i] = mat[i][j];
				}
			}
			
			//Vector
			for (int i=0; i<nDim; i++) {
				coeff[nDim-1-i] += y[k]*xps[xps.length-1-i];
			}
		}
		
		//Solve
		//gjSolve(mat, coeff);
		choleskySolve(mat, coeff);
	}
	private static void calcPows(float[] xp, float x) {
		xp[xp.length-1] = 1f;
		for (int i=1; i<xp.length; i++) {
			xp[xp.length-1-i] = x*xp[xp.length-i];
		}
	}
/*	
	private static void gjSolve(double[][] mat, float[] vect) {
		for (int i=0; i<vect.length; i++) {
			double piv = mat[i][i];
			
			for (int j=i+1; j<vect.length; j++) mat[i][j] = mat[i][j]/piv;
			vect[i] = (float)(vect[i]/piv);
			
			for (int j=0; j<i; j++) {
				double mult = -mat[j][i];
				for (int k=i+1; k<vect.length; k++) mat[j][k] += mult*mat[i][k];
				vect[j] += (float)(mult*vect[i]);
			}
			for (int j=i+1; j<vect.length; j++) {
				double mult = -mat[j][i];
				for (int k=i+1; k<vect.length; k++) mat[j][k] += mult*mat[i][k];
				vect[j] += (float)(mult*vect[i]);
			}
		}
	}
*/	
	private static void choleskySolve(double[][] mat, float[] vect) {
		//Copy
		double[] vectd = new double[vect.length];
		for (int i=0; i<vect.length; i++) vectd[i] = vect[i];
		
		//Decompose
		choleskyDecompose(mat);

		//Solve L * v = y
		for (int i=0; i<vect.length; i++) {
			double sumPrev = 0f;
			for (int j=0; j<i; j++) sumPrev += mat[i][j]*vectd[j];
			vectd[i] = (vectd[i]-sumPrev)/mat[i][i];
		}
		
		//Solve L' * x = v
		for (int i=vect.length-1; i>=0; i--) {
			double sumRH = 0f;
			for (int j=i+1; j<vect.length; j++) sumRH += mat[i][j]*vectd[j];
			vectd[i] = (vectd[i]-sumRH)/mat[i][i];
		}
		
		//Return
		for (int i=0; i<vect.length; i++) vect[i] = (float)vectd[i];
	}
	private static void choleskyDecompose(double[][] mat) {
		int n = mat.length;
		for (int i=0; i<n; i++) {
			//Square root of the diagonal element
			mat[i][i] = Math.sqrt(mat[i][i]);
			
			//Same row/column
			for (int j=i+1; j<n; j++) {
				mat[j][i] = mat[j][i]/mat[i][i];
				mat[i][j] = mat[j][i];
			}
			
			//Sub-matrix
			for (int j=i+1; j<n; j++) {
				mat[j][j] = mat[j][j] - mat[i][j]*mat[i][j];
				for (int k=j+1; k<n; k++) {
					mat[j][k] = mat[j][k] - mat[i][j]*mat[i][k];
					mat[k][j] = mat[j][k];
				}
			}
		}
	}
}
