/*
Copyright 2018 Toyota Research Institute North America


Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation and/or
 other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors 
may be used to endorse or promote products derived from this software without 
specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package fastsimj2.components;

import fastsimj2.*;

public class FuelConverterModel {
	private float maxOutPowerKW;
	private float powerAtMaxEffKW;
	private float[] aCoeff;
	
	public float inputPowerKW(float outPowerKW) {
		float x = outPowerKW;
		if (x < 0) x = 0;
		if (x > maxOutPowerKW) x = maxOutPowerKW;
		
		float inPow = 0f;
		float xp = 1f;
		
		for (int i=0; i<aCoeff.length; i++) {
			inPow += xp * aCoeff[aCoeff.length-1-i];
			xp *= x;
		}
		
		return inPow;
	}
	public float powerAtMaxEffKW() {return powerAtMaxEffKW;}
	
	private static int polyOrder() {return 4;}
	private static int nFitPoints() {return 21;}
	
	public FuelConverterModel(FSJVehicleModelParameters vehModel) {
		maxOutPowerKW = vehModel.maxFuelConvKw;
		
		float[] x4_SI = {0.0f, 0.1f, 0.28f, 1.0f};
		float[] x4_AT = {0.0f, 0.1f, 0.4f, 1.0f};
		float[] x4_DZ = {0.0f, 0.1f, 0.4f, 1.0f};
		
		float[] y4_SI = {0.1f, 0.28f, 0.36f, 0.26f};
		float[] y4_AT = {0.08f, 0.36f, 0.36f, 0.31f};
		float[] y4_DZ = {0.12f, 0.31f, 0.36f, 0.33f};
		
		float[] x = new float[nFitPoints()];
		float[] y = new float[nFitPoints()];
		
		switch (vehModel.fcEffType) {
		case atkinson:
			prepCoeff(x,y,x4_AT,y4_AT);
			break;
		case diesel:
			prepCoeff(x,y,x4_DZ,y4_DZ);
			break;
		case sparkIgnition:
			prepCoeff(x,y,x4_SI,y4_SI);
			break;
		}

		calcCoeff(x,y);
	}


	private void prepCoeff(float[] x, float[] y, float[] x4, float[] y4) {
		int nDiv = x.length-1;
		for (int i=0; i <= nDiv; i++) {
			float z = (float)i/(float)nDiv;
			x[i] = z*maxOutPowerKW;
			y[i] = x[i]/linInterpolate(x4,y4,z);
		}
	}
	private static float linInterpolate(float[] xs, float[] ys, float x) {
		if (x <= xs[0]) return ys[0];
		if (x >= xs[xs.length-1]) return ys[xs.length-1];
		
		int intervalID = 0;
		while (xs[intervalID+1] < x) intervalID++;
		
		float x1 = xs[intervalID];
		float x2 = xs[intervalID+1];
		float z = (x-x1)/(x2-x1);
		
		return ys[intervalID] + z*(ys[intervalID+1]-ys[intervalID]);
	}
	private void calcCoeff(float[] x, float[] y) {
		int nDim = polyOrder() + 1;
		if (nDim > x.length) nDim = x.length;
		
		aCoeff = new float[nDim];
		float[] xps = new float[2*polyOrder() + 1];
		double[][] mat = new double[nDim][nDim];
		
		for (int k=0; k<x.length; k++) {	//Loop on points
			calcPows(xps, x[k]);	//Powers of x
			
			//Matrix coefficients
			for (int i=0; i<nDim; i++) {
				mat[i][i] += xps[2*i];
				
				for (int j=i+1; j<nDim; j++) {
					mat[i][j] += xps[i+j];
					mat[j][i] = mat[i][j];
				}
			}
			
			//Vector
			for (int i=0; i<nDim; i++) {
				aCoeff[nDim-1-i] += y[k]*xps[xps.length-1-i];
			}
		}
		
		//Solve
		gjSolve(mat, aCoeff);
		
		//Identify output power at maximum efficiency
		float maxModelEff = 0f;
		powerAtMaxEffKW = 0f;
		
		for (int i=1; i<x.length; i++) {
			float curEff = x[i]/inputPowerKW(x[i]);
			if (curEff > maxModelEff) {
				maxModelEff = curEff;
				powerAtMaxEffKW = x[i];
			}
		}
	}
	private static void calcPows(float[] xp, float x) {
		xp[xp.length-1] = 1f;
		for (int i=1; i<xp.length; i++) {
			xp[xp.length-1-i] = x*xp[xp.length-i];
		}
	}
	private static void gjSolve(double[][] mat, float[] vect) {
		for (int i=0; i<vect.length; i++) {
			double piv = mat[i][i];
			
			for (int j=i+1; j<vect.length; j++) mat[i][j] = mat[i][j]/piv;
			vect[i] = (float)(vect[i]/piv);
			
			for (int j=0; j<i; j++) {
				double mult = -mat[j][i];
				for (int k=i+1; k<vect.length; k++) mat[j][k] += mult*mat[i][k];
				vect[j] += (float)(mult*vect[i]);
			}
			for (int j=i+1; j<vect.length; j++) {
				double mult = -mat[j][i];
				for (int k=i+1; k<vect.length; k++) mat[j][k] += mult*mat[i][k];
				vect[j] += (float)(mult*vect[i]);
			}
		}
	}
}
