/*
Copyright 2018 Toyota Research Institute North America


Redistribution and use in source and binary forms, with or without modification,
 are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation and/or
 other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors 
may be used to endorse or promote products derived from this software without 
specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

package fastsimj2.components;

import java.io.FileWriter;
import java.io.IOException;

import fastsimj2.*;

public class VehicleState {
	
	//Link to Simulation constants
	private FSJSimConstants simConsts;
	//Link to vehicle model parameters
	private FSJVehicleModelParameters vehModel;
	public FSJVehicleModelParameters vehModel() {return vehModel;}
	//Link to Fuel converter model
	private FuelConverterModel fcModel;
	//Link to Electric motor model
	private ElectMotorModel mtModel;
	
	
	//Public Quantity values
	public float time,speedMPH,speedMS,roadGrade, distSLSMeters,distSLSMiles,
		dragKW, ascentKW, rollResKW, accelKW, fricBrakeKW,
		fcPowerOut,fcPowerIn,fcEffn, fuelUseSLS,
		mtPowerOut,mtPowerIn,mtEffn,regenKW, batteryUseSLS, relSoC,absSoC, speedTraceSlipMPH;
	
	//Private Quantity values
	private float totalVehMass, allWheelsInertia, maxAccelMS2;
	private boolean highAccFCOn;
	
	
	//Constructor
	public VehicleState(FSJSimConstants simulationConsts, FSJVehicleModelParameters vehicleModel, 
			FuelConverterModel fuelConvModel, ElectMotorModel motorModel) {
		//Link objects
		simConsts = simulationConsts;
		vehModel = vehicleModel;
		fcModel = fuelConvModel;
		mtModel = motorModel;
		
		//Pre-calculate mass, inertia and some other constants
		allWheelsInertia = vehModel.numWheels*vehModel.wheelInteriaKgM2;
		float gliderMass = vehModel.gliderKg;
		float cargoMass = vehModel.cargoKg;
		float transMass = vehModel.transKg;		
		float tankMass = 0f;
		float fcMass = 0f;
		float essMass = 0f;
		float mtMass = 0f;
		switch (vehModel.vehPtType) {
		case bev:
			essMass = vehModel.essBaseKg + vehModel.maxEssKwh * vehModel.essKgPerKwh;
			mtMass = vehModel.mcPeBaseKg + vehModel.maxMotorKw * vehModel.mcPeKgPerKw;
			break;
		case cv:
			tankMass = vehModel.fuelStorKwh / vehModel.fuelStorKwhPerKg;
			fcMass = vehModel.maxFuelConvKw / vehModel.fuelConvKwPerKg;
			break;
		case hev:
		case phev:
			essMass = vehModel.essBaseKg + vehModel.maxEssKwh * vehModel.essKgPerKwh;
			mtMass = vehModel.mcPeBaseKg + vehModel.maxMotorKw * vehModel.mcPeKgPerKw;
			tankMass = vehModel.fuelStorKwh / vehModel.fuelStorKwhPerKg;
			fcMass = vehModel.maxFuelConvKw / vehModel.fuelConvKwPerKg;
			break;
		}
		totalVehMass = gliderMass + cargoMass + tankMass + fcMass + mtMass + essMass + transMass;
		maxAccelMS2 = vehModel.wheelCoefOfFric*vehModel.driveAxleWeightFrac*simConsts.gravity/
				(1f + (vehModel.vehCgM*vehModel.wheelCoefOfFric/vehModel.wheelBaseM));
		highAccFCOn = false;
		
		
		resetAllButSoC(0f, 0f, 0f);
		
		switch (vehModel.vehPtType) {
		case bev:
		case phev:
			setRelSoC(1f);
			break;
		case hev:
			setRelSoC(0f);
			break;
		default:
			setRelSoC(0f);
			break;		
		}
	}
	//Copy constructor
	public VehicleState(VehicleState other) {
		time = other.time;
		speedMPH = other.speedMPH;
		speedMS = other.speedMS;
		roadGrade = other.roadGrade;
		
		distSLSMeters = other.distSLSMeters;
		distSLSMiles = other.distSLSMiles;
		
		dragKW = other.dragKW;
		ascentKW = other.ascentKW;
		rollResKW = other.rollResKW;
		accelKW = other.accelKW;
		fricBrakeKW = other.fricBrakeKW;
		fcPowerOut = other.fcPowerOut;
		fcPowerIn = other.fcPowerIn;
		fcEffn = other.fcEffn;
		fuelUseSLS = other.fuelUseSLS;
		mtPowerOut = other.mtPowerOut;
		mtPowerIn = other.mtPowerIn;
		mtEffn = other.mtEffn;
		regenKW = other.regenKW;
		batteryUseSLS = other.batteryUseSLS;
		relSoC = other.relSoC;
		absSoC = other.absSoC;
		speedTraceSlipMPH = other.speedTraceSlipMPH;
		
		totalVehMass = other.totalVehMass;
		allWheelsInertia = other.allWheelsInertia;
		maxAccelMS2 = other.maxAccelMS2;
		highAccFCOn = other.highAccFCOn;
	}
	//Function to reset all values except battery state of charge values
	public void resetAllButSoC(float t, float mph, float grade) {
		time = t;
		speedMPH = mph;
		speedMS = speedMPH/simConsts.mphPerMps;
		roadGrade = grade;
		
		distSLSMeters = 0f;
		distSLSMiles = 0f;
		
		dragKW = 0f;
		ascentKW = 0f;
		rollResKW = 0f;
		accelKW = 0f;
		fricBrakeKW = 0f;
		fcPowerOut = 0f;
		fcPowerIn = 0f;
		fcEffn = 0f;
		fuelUseSLS = 0f;
		mtPowerOut = 0f;
		mtPowerIn = 0f;
		mtEffn = 0f;
		regenKW = 0f;
		batteryUseSLS = 0f;
		speedTraceSlipMPH = 0f;
		
		highAccFCOn = false;
	}
	//Function to set the state of charge
	public void setRelSoC(float rSoC) {
		float deltaAbsSoC = vehModel.maxSoc - vehModel.minSoc;
		relSoC = rSoC;
		absSoC = vehModel.minSoc + rSoC*deltaAbsSoC;
	}
	public void setAbsSoC(float aSoC) {
		absSoC = aSoC;
		float deltaAbsSoC = vehModel.maxSoc - vehModel.minSoc;
		if (deltaAbsSoC > 0) relSoC = (absSoC-vehModel.minSoc)/deltaAbsSoC;
		else relSoC = 1f;
	}

	//Function to update to next time step
	public void updateState(float nextTime, float nextSpeedMPH, float nextRoadGrade, float addAuxPower) {
		switch (vehModel.vehPtType) {
		case bev:
			updateState_BEV(nextTime, nextSpeedMPH, nextRoadGrade, addAuxPower);
			break;
		case cv:
			updateState_CV(nextTime, nextSpeedMPH, nextRoadGrade, addAuxPower);
			break;
		case hev:
		case phev:
			updateState_HEV_PHEV(nextTime, nextSpeedMPH, nextRoadGrade, addAuxPower);
			break;
		}
	}
	//BEV
	private void updateState_BEV(float nextTime, float nextSpeedMPH, float nextRoadGrade, float addAuxPower) {
		//Time step size
		float deltaTime = cDeltaTime(time, nextTime);
		
		//Target speed
		float nextSpeedMS = cCovSpeedMPHtoPMS(nextSpeedMPH);
		float targetSpeedMS = cTargetSpeed(nextSpeedMS, deltaTime);
		
		//Limits on motor power
		float curMaxElectKW = cMaxElectKW(deltaTime, addAuxPower, false);	//Do the simulation even if running below battery capacity
		float curMaxMcKwOut = cMaxMotorMcKWOut(deltaTime, curMaxElectKW);
		float essCapChargeLim = cEssCapChargeLim(deltaTime);
		float curMaxMechMcKwIn = cMaxMotoMcKwIn(essCapChargeLim);
		
		//Limit on mechanical power transmitted to wheels
		float limMechKWTransToWheels = (curMaxMcKwOut - vehModel.auxKw - addAuxPower)*vehModel.transEff;
		
		//Power Consumption
		dragKW = cDragKW(0.5f*(speedMS+targetSpeedMS));
		accelKW = cAccelKW(targetSpeedMS, deltaTime);
		ascentKW = cAscentKW(nextRoadGrade, 0.5f*(speedMS+targetSpeedMS));
		float wheelInertiaKW = cWheelInertiaKW(targetSpeedMS, deltaTime);
		rollResKW = cRollResKW(0.5f*(speedMS+targetSpeedMS));
		
		float sumKWReqToWheel = dragKW + accelKW + ascentKW + wheelInertiaKW + rollResKW;
		float contrLimPercentRegen = maxContrLimPercentRegen(vehModel.maxRegen, 0.5f*(speedMPH+targetSpeedMS*simConsts.mphPerMps));
		regenKW = Math.max(0, Math.min(curMaxMechMcKwIn, -sumKWReqToWheel*contrLimPercentRegen));
		fricBrakeKW = -Math.min(0, sumKWReqToWheel+regenKW);
		float transKwOutReq = sumKWReqToWheel + fricBrakeKW;

		//Adjust attainable speed if can't output enough power
		float adjustSpeedMSStep = 0.01f;
		float achievedSpeedMS = targetSpeedMS;
		float kWShortage = limMechKWTransToWheels - transKwOutReq;
		
		while ((kWShortage < 0)&&(achievedSpeedMS>adjustSpeedMSStep)) {
			achievedSpeedMS += -adjustSpeedMSStep;

			dragKW = cDragKW(0.5f*(speedMS+achievedSpeedMS));
			accelKW = cAccelKW(achievedSpeedMS, deltaTime);
			ascentKW = cAscentKW(nextRoadGrade, 0.5f*(speedMS+achievedSpeedMS));
			wheelInertiaKW = cWheelInertiaKW(achievedSpeedMS, deltaTime);
			rollResKW = cRollResKW(0.5f*(speedMS+achievedSpeedMS));
			
			sumKWReqToWheel = dragKW + accelKW + ascentKW + wheelInertiaKW + rollResKW;
			contrLimPercentRegen = maxContrLimPercentRegen(vehModel.maxRegen, 0.5f*(speedMPH+achievedSpeedMS*simConsts.mphPerMps));
			regenKW = Math.max(0, Math.min(curMaxMechMcKwIn, -sumKWReqToWheel*contrLimPercentRegen));
			fricBrakeKW = -Math.min(0, sumKWReqToWheel+regenKW);
			transKwOutReq = sumKWReqToWheel + fricBrakeKW;
			kWShortage = limMechKWTransToWheels - transKwOutReq;
		}
		
		
		//Update speed, distance and time
		distSLSMeters = 0.5f*deltaTime*(achievedSpeedMS+speedMS);
		distSLSMiles = distSLSMeters/simConsts.metersPerMile;
		
		speedMS = achievedSpeedMS;
		speedMPH = speedMS*simConsts.mphPerMps;
		time = nextTime;
		
		//Update energy usage
		float achievedTransPowerInKW = 0f;
		if (transKwOutReq>=0) achievedTransPowerInKW = transKwOutReq/vehModel.transEff;
		else achievedTransPowerInKW = transKwOutReq*vehModel.transEff;
		
		float achievedMotorKWOut = achievedTransPowerInKW;
		if (achievedTransPowerInKW<0) achievedMotorKWOut = -Math.min(curMaxMechMcKwIn, -achievedTransPowerInKW);

		float mcElecKwInAch = 0f;	
		if (achievedMotorKWOut < 0f) {
			mcElecKwInAch = -Math.max(mtModel.outputPowerKW(-achievedMotorKWOut),
					-mtModel.minMotorEff()*achievedMotorKWOut);
		}
		else if (achievedMotorKWOut > 0) {
			mcElecKwInAch = mtModel.inputPowerKW(achievedMotorKWOut);
		}
		
		float essKwOutAch = mcElecKwInAch + vehModel.auxKw + addAuxPower;
		
		mtPowerOut = achievedMotorKWOut;
		mtPowerIn = mcElecKwInAch;
		mtEffn = 0f;
		if (mtPowerIn>0) mtEffn = mtPowerOut/mtPowerIn;
		if (mtPowerIn<0) mtEffn = mtPowerIn/mtPowerOut;
		
		regenKW = 0f;
		if (mtPowerIn<0) regenKW = -mtPowerIn;
		
		batteryUseSLS = essKwOutAch*deltaTime/3600.0f;
		if (batteryUseSLS < 0) batteryUseSLS *= (float)Math.sqrt(vehModel.essRountTripEff);
		else batteryUseSLS = batteryUseSLS/(float)Math.sqrt(vehModel.essRountTripEff);
		
		float deltaAbsSoC = batteryUseSLS/vehModel.maxEssKwh;
		absSoC += -deltaAbsSoC;
		relSoC = (absSoC-vehModel.minSoc)/(vehModel.maxSoc-vehModel.minSoc);

		speedTraceSlipMPH = nextSpeedMPH - speedMPH;
	}
	//HEV and PHEV
	private void updateState_HEV_PHEV(float nextTime, float nextSpeedMPH, float nextRoadGrade, float addAuxPower) {
		//Time step size
		float deltaTime = cDeltaTime(time, nextTime);
		
		//Target speed
		float nextSpeedMS = cCovSpeedMPHtoPMS(nextSpeedMPH);
		float targetSpeedMS = cTargetSpeed(nextSpeedMS, deltaTime);
		
		//Limit on ICE power
		float limFCOut = cLimFCOut(deltaTime);
		
		//Limits on motor power
		float curMaxElectKW = cMaxElectKW(deltaTime, addAuxPower, true);
		float curMaxMcKwOut = cMaxMotorMcKWOut(deltaTime, curMaxElectKW);
		float essCapChargeLim = cEssCapChargeLim(deltaTime);
		float curMaxMechMcKwIn = cMaxMotoMcKwIn(essCapChargeLim);
		
		//Fuel converter flag
		if (highAccFCOn) {
			if (absSoC < (vehModel.minSoc+vehModel.percHighAccBuf)) highAccFCOn = true;
			else highAccFCOn = false;
		}
		if (absSoC < vehModel.minSoc) highAccFCOn = true;

		//Limit on mechanical power transmitted to wheels
		float limMechKWTransToWheels = limFCOut + curMaxMcKwOut;
		if (highAccFCOn) limMechKWTransToWheels += -vehModel.auxKw - addAuxPower;
		limMechKWTransToWheels += Math.min(0, curMaxElectKW);
		limMechKWTransToWheels = limMechKWTransToWheels*vehModel.transEff;
		
		//Power Consumption
		dragKW = cDragKW(0.5f*(speedMS+targetSpeedMS));
		accelKW = cAccelKW(targetSpeedMS, deltaTime);
		ascentKW = cAscentKW(nextRoadGrade, 0.5f*(speedMS+targetSpeedMS));
		float wheelInertiaKW = cWheelInertiaKW(targetSpeedMS, deltaTime);
		rollResKW = cRollResKW(0.5f*(speedMS+targetSpeedMS));
		
		float sumKWReqToWheel = dragKW + accelKW + ascentKW + wheelInertiaKW + rollResKW;
		float contrLimPercentRegen = maxContrLimPercentRegen(vehModel.maxRegen, 0.5f*(speedMPH+targetSpeedMS*simConsts.mphPerMps));
		regenKW = Math.max(0, Math.min(curMaxMechMcKwIn, -sumKWReqToWheel*contrLimPercentRegen));
		fricBrakeKW = -Math.min(0, sumKWReqToWheel+regenKW);
		float transKwOutReq = sumKWReqToWheel + fricBrakeKW;
		
		//Adjust attainable speed if can't output enough power
		float adjustSpeedMSStep = 0.01f;
		float achievedSpeedMS = targetSpeedMS;
		float kWShortage = limMechKWTransToWheels - transKwOutReq;
		
		while ((kWShortage < 0)&&(achievedSpeedMS>adjustSpeedMSStep)) {
			achievedSpeedMS += -adjustSpeedMSStep;

			dragKW = cDragKW(0.5f*(speedMS+achievedSpeedMS));
			accelKW = cAccelKW(achievedSpeedMS, deltaTime);
			ascentKW = cAscentKW(nextRoadGrade, 0.5f*(speedMS+achievedSpeedMS));
			wheelInertiaKW = cWheelInertiaKW(achievedSpeedMS, deltaTime);
			rollResKW = cRollResKW(0.5f*(speedMS+achievedSpeedMS));
			
			sumKWReqToWheel = dragKW + accelKW + ascentKW + wheelInertiaKW + rollResKW;
			contrLimPercentRegen = maxContrLimPercentRegen(vehModel.maxRegen, 0.5f*(speedMPH+achievedSpeedMS*simConsts.mphPerMps));
			regenKW = Math.max(0, Math.min(curMaxMechMcKwIn, -sumKWReqToWheel*contrLimPercentRegen));
			fricBrakeKW = -Math.min(0, sumKWReqToWheel+regenKW);
			transKwOutReq = sumKWReqToWheel + fricBrakeKW;
			kWShortage = limMechKWTransToWheels - transKwOutReq;
		}
		
		//Update speed, distance and time
		distSLSMeters = 0.5f*deltaTime*(achievedSpeedMS+speedMS);
		distSLSMiles = distSLSMeters/simConsts.metersPerMile;
		
		speedMS = achievedSpeedMS;
		speedMPH = speedMS*simConsts.mphPerMps;
		time = nextTime;
		
		//Update energy usage
		float achievedTransPowerInKW = 0f;
		if (transKwOutReq>=0) achievedTransPowerInKW = transKwOutReq/vehModel.transEff;
		else achievedTransPowerInKW = transKwOutReq*vehModel.transEff;
		
		float achievedMotorKWOut = cHybridMotorKWOut(achievedTransPowerInKW, curMaxMechMcKwIn, addAuxPower, curMaxElectKW, essCapChargeLim, curMaxMcKwOut, limFCOut);
		
		float mcElecKwInAch = 0f;	
		if (achievedMotorKWOut < 0f) {
			mcElecKwInAch = -Math.max(mtModel.outputPowerKW(-achievedMotorKWOut),
					-mtModel.minMotorEff()*achievedMotorKWOut);
		}
		else if (achievedMotorKWOut > 0) {
			mcElecKwInAch = mtModel.inputPowerKW(achievedMotorKWOut);
		}
		
		float essKwOutAch = mcElecKwInAch;
		if (!highAccFCOn) essKwOutAch += vehModel.auxKw + addAuxPower;
		
		float fcKWOutAch = achievedTransPowerInKW - achievedMotorKWOut;
		if (highAccFCOn) fcKWOutAch += vehModel.auxKw + addAuxPower;
		
		mtPowerOut = achievedMotorKWOut;
		mtPowerIn = mcElecKwInAch;
		mtEffn = 0f;
		if (mtPowerIn>0) mtEffn = mtPowerOut/mtPowerIn;
		if (mtPowerIn<0) mtEffn = mtPowerIn/mtPowerOut;
		
		regenKW = 0f;
		if (mtPowerIn<0) regenKW = -mtPowerIn;
		
		batteryUseSLS = essKwOutAch*deltaTime/3600.0f;
		if (batteryUseSLS < 0) batteryUseSLS *= (float)Math.sqrt(vehModel.essRountTripEff);
		else batteryUseSLS = batteryUseSLS/(float)Math.sqrt(vehModel.essRountTripEff);
		
		float deltaAbsSoC = batteryUseSLS/vehModel.maxEssKwh;
		absSoC += -deltaAbsSoC;
		relSoC = (absSoC-vehModel.minSoc)/(vehModel.maxSoc-vehModel.minSoc);
		
		fcPowerOut = fcKWOutAch;
		fcPowerIn = 0f;
		if (fcPowerOut > 0) fcPowerIn = fcModel.inputPowerKW(fcPowerOut);
		fcEffn = 0f;
		if (fcPowerIn > 0) fcEffn = fcPowerOut/fcPowerIn;
		
		float kwhIn = fcPowerIn*deltaTime/3600.0f;
		if (vehModel.fcEffType == FuelConverterEffType.diesel) fuelUseSLS = kwhIn / simConsts.kWhPerGalDiesel;
		else fuelUseSLS = kwhIn / simConsts.kWhPerGGE;

		speedTraceSlipMPH = nextSpeedMPH - speedMPH;
	}
	//CV
	private void updateState_CV(float nextTime, float nextSpeedMPH, float nextRoadGrade, float addAuxPower) {
		//Time step size
		float deltaTime = cDeltaTime(time, nextTime);
		
		//Target speed
		float nextSpeedMS = cCovSpeedMPHtoPMS(nextSpeedMPH);
		float targetSpeedMS = cTargetSpeed(nextSpeedMS, deltaTime);
		
		//Limit on ICE power
		float limFCOut = cLimFCOut(deltaTime);
		float limMechKWTransToWheels = (limFCOut-vehModel.auxKw-addAuxPower)*vehModel.transEff;
		
		//Power Consumption
		dragKW = cDragKW(0.5f*(speedMS+targetSpeedMS));
		accelKW = cAccelKW(targetSpeedMS, deltaTime);
		ascentKW = cAscentKW(nextRoadGrade, 0.5f*(speedMS+targetSpeedMS));
		float wheelInertiaKW = cWheelInertiaKW(targetSpeedMS, deltaTime);
		rollResKW = cRollResKW(0.5f*(speedMS+targetSpeedMS));
		
		float sumKWReqToWheel = dragKW + accelKW + ascentKW + wheelInertiaKW + rollResKW;
		fricBrakeKW = -Math.min(0, sumKWReqToWheel);
		
		//Adjust attainable speed if can't output enough power
		float adjustSpeedMSStep = 0.01f;
		float adjustedSpeedMS = targetSpeedMS;
		float kWShortage = limMechKWTransToWheels - (sumKWReqToWheel+fricBrakeKW);
		
		while ((kWShortage < 0)&&(adjustedSpeedMS>adjustSpeedMSStep)) {
			adjustedSpeedMS += -adjustSpeedMSStep;
			
			dragKW = cDragKW(0.5f*(speedMS+adjustedSpeedMS));
			accelKW = cAccelKW(adjustedSpeedMS, deltaTime);
			ascentKW = cAscentKW(nextRoadGrade, 0.5f*(speedMS+adjustedSpeedMS));
			wheelInertiaKW = cWheelInertiaKW(adjustedSpeedMS, deltaTime);
			rollResKW = cRollResKW(0.5f*(speedMS+adjustedSpeedMS));
			
			sumKWReqToWheel = dragKW + accelKW + ascentKW + wheelInertiaKW + rollResKW;
			fricBrakeKW = -Math.min(0, sumKWReqToWheel);
			kWShortage = limMechKWTransToWheels - (sumKWReqToWheel+fricBrakeKW);
		}

		//Update state
		distSLSMeters = 0.5f*deltaTime*(adjustedSpeedMS+speedMS);
		distSLSMiles = distSLSMeters/simConsts.metersPerMile;
		
		speedMS = adjustedSpeedMS;
		speedMPH = speedMS*simConsts.mphPerMps;
		time = nextTime;
		
		fcPowerOut = (sumKWReqToWheel+fricBrakeKW)/vehModel.transEff + vehModel.auxKw + addAuxPower;
		fcPowerIn = fcModel.inputPowerKW(fcPowerOut);
		fcEffn = fcPowerOut/fcPowerIn;
		
		float kwhIn = fcPowerIn*deltaTime/3600.0f;
		if (vehModel.fcEffType == FuelConverterEffType.diesel) fuelUseSLS = kwhIn / simConsts.kWhPerGalDiesel;
		else fuelUseSLS = kwhIn / simConsts.kWhPerGGE;
		
		speedTraceSlipMPH = nextSpeedMPH - speedMPH;
	}
	
	//Function for calculating difference in time
	private float cDeltaTime(float curTime, float nextTime) {
		return nextTime - curTime;
	}
	//Function for converting speed from MPH to meters per second
	private float cCovSpeedMPHtoPMS(float mph) {
		return mph/simConsts.mphPerMps;
	}
	//Function for calculating target speed (as dictated by acceleration limits)
	private float cTargetSpeed(float nextSpeedMPS, float deltaTime) {
		if (nextSpeedMPS<0) nextSpeedMPS = 0;
		float maxDeltaSpeedMS = maxAccelMS2*deltaTime;
		float reqSpeedMS = nextSpeedMPS;
		if (reqSpeedMS > (speedMS+maxDeltaSpeedMS)) reqSpeedMS = speedMS+maxDeltaSpeedMS;
		if (reqSpeedMS < (speedMS-maxDeltaSpeedMS)) reqSpeedMS = speedMS-maxDeltaSpeedMS;
		return reqSpeedMS;
	}
	//Function for calculating maximum power from Fuel converter
	private float cLimFCOut(float deltaTime) {
		float limFuelStoreKW = Math.min(vehModel.maxFuelStorKw, 
				fcPowerOut+vehModel.maxFuelStorKw*(deltaTime/(float)vehModel.fuelStorSecsToPeakPwr));
		float limFCOut = Math.min(vehModel.maxFuelConvKw, 
				fcPowerOut+vehModel.maxFuelConvKw*(deltaTime/(float)vehModel.fuelConvSecsToPeakPwr));
		return Math.min(limFCOut, limFuelStoreKW);
	}
	//Function for calculating drag power
	private float cDragKW(float speed) {
		return 0.5f*simConsts.airDensity*vehModel.dragCoef*vehModel.frontalAreaM2*(float)Math.pow(speed, 3f)/1000f;
	}
	//Function for calculating acceleration power
	private float cAccelKW(float targetSpeedMS, float deltaTime) {
		return 0.5f*totalVehMass*(targetSpeedMS*targetSpeedMS - speedMS*speedMS)/(deltaTime*1000f);
	}
	//Function for calculating road grade power
	private float cAscentKW(float nextRoadGrade, float speed) {
		return totalVehMass*simConsts.gravity*(float)Math.sin(Math.atan(nextRoadGrade))*speed/1000f;
	}
	//Function for calculating wheel inertia power
	private float cWheelInertiaKW(float targetSpeedMS, float deltaTime) {
		float curWheelRadPS = speedMS/vehModel.wheelRadiusM;
		float reqWheelRadPS = targetSpeedMS/vehModel.wheelRadiusM;
		return 0.5f*allWheelsInertia*(reqWheelRadPS*reqWheelRadPS - curWheelRadPS*curWheelRadPS)/(deltaTime*1000f);		
	}
	//Function for calculating rolling resistance power
	private float cRollResKW(float speed) {
		return simConsts.gravity*totalVehMass*vehModel.wheelRrCoef*speed/1000f;
	}
	//Function for calculating maximum electric power from discharge
	private float cMaxElectKW(float deltaTime, float aaux, boolean useMinSoCLimit) {	
		float essCapDischargeLim = vehModel.maxEssKw;
		if (vehModel.useMotorKwInsteadOfESS) essCapDischargeLim = vehModel.maxMotorKw/vehModel.motorPeakEff;
		if (useMinSoCLimit) {
			essCapDischargeLim = Math.min(vehModel.maxEssKwh*(float)Math.sqrt(vehModel.essRountTripEff)*
					3600f*(absSoC-vehModel.minSoc)/deltaTime, essCapDischargeLim);
			if (absSoC < vehModel.minSoc) essCapDischargeLim = 0f;
		}
		else {
			essCapDischargeLim = Math.min(vehModel.maxEssKwh*(float)Math.sqrt(vehModel.essRountTripEff)*
					3600f*(1f-vehModel.minSoc)/deltaTime, essCapDischargeLim);
		}
		return essCapDischargeLim - vehModel.auxKw - aaux;
	}
	//Function for calculating maximum motor mechanical power output
	private float cMaxMotorMcKWOut(float deltaTime, float curMaxElectKW) {
		float curMaxMcKwOut = 0f;
		if (curMaxElectKW > 0) curMaxMcKwOut = mtModel.outputPowerKW(curMaxElectKW);
		curMaxMcKwOut = Math.min(curMaxMcKwOut, Math.abs(mtPowerOut)+deltaTime*vehModel.maxMotorKw/vehModel.motorSecsToPeakPwr);
		return Math.max(curMaxMcKwOut, -vehModel.maxMotorKw);
	}
	//Function for calculating energy charging capacity limit
	private float cEssCapChargeLim(float deltaTime) {
		float essCapDischargeLim = vehModel.maxEssKw;
		if (vehModel.useMotorKwInsteadOfESS) essCapDischargeLim = vehModel.maxMotorKw/vehModel.motorPeakEff;
		float essCapChargeLim = Math.max(vehModel.maxEssKwh*(vehModel.maxSoc-absSoC)*3600f/
				((float)Math.sqrt(vehModel.essRountTripEff)*deltaTime),0);
		return Math.min(essCapChargeLim, essCapDischargeLim);
	}
	//Function for calculating maximum mechanical power to be input to the motor (when charging)
	private float cMaxMotoMcKwIn(float essCapChargeLim) {		
		float essLimMcRegenKw = 0f;
		if (essCapChargeLim > 0) essLimMcRegenKw = mtModel.inputPowerKW(essCapChargeLim);
		return Math.min(essLimMcRegenKw, vehModel.maxMotorKw);
	}
	
	//Function for calculation of percentage regenerative breaking
	private static float maxContrLimPercentRegen(float maxRegenFrac, float mph) {
		float a = 1000f;
		float b = 0.9f;
		return maxRegenFrac/(1f+a*(float)Math.exp(-b*(1f+mph)));
	}

	//Function for calculating hybrid decision making for motor power out
	private float cHybridMotorKWOut(float achievedTransPowerInKW, float curMaxMechMcKwIn, float addAuxPower, 
			float curMaxElectKW, float essCapChargeLim, float curMaxMcKwOut, float limFCOut) {
		float achievedMotorKWOut = 0f;
		
		if (achievedTransPowerInKW<0) achievedMotorKWOut = -Math.min(curMaxMechMcKwIn, -achievedTransPowerInKW);
		else {
			float essDesiredDischgKw;
			float mcKwReqFromEss4AE = 0f;
			if (achievedTransPowerInKW > 0) mcKwReqFromEss4AE = mtModel.inputPowerKW(achievedTransPowerInKW);
			float allElectEssKW = mcKwReqFromEss4AE + vehModel.auxKw + addAuxPower;
						
			if ((speedMPH>vehModel.mphFcOn)||(allElectEssKW>=vehModel.kwDemandFcOn)||(highAccFCOn)) 
			{
				switch (vehModel.controlStrategy) {
				case engineEfficiency:
					float tPwr = achievedTransPowerInKW + vehModel.auxKw + addAuxPower;
					float fcEffDesiredDischgKw = 0f;
					
					if ((allElectEssKW <= curMaxElectKW)
						&&(allElectEssKW <= vehModel.percFcPwrDoneElec*vehModel.maxFuelConvKw)
						&&(allElectEssKW > 0)) 
					{
						fcEffDesiredDischgKw = allElectEssKW;
					}
					else if ((tPwr < fcModel.powerAtMaxEffKW())
							&&(tPwr > 0)
							&&(vehModel.vehPtType == VehicleDriveTrainType.hev)) 
					{
						fcEffDesiredDischgKw = Math.max(-essCapChargeLim, -fcModel.powerAtMaxEffKW()+tPwr);
					}
					else if ((tPwr > fcModel.powerAtMaxEffKW())
							&&(tPwr > 0)) 
					{
						fcEffDesiredDischgKw = Math.min(curMaxElectKW, tPwr-fcModel.powerAtMaxEffKW());
					}
	
					essDesiredDischgKw = fcEffDesiredDischgKw;
					break;
				default:	//kineticEnergy
					float percToLowSocMph = Math.min(1f, 
							(speedMPH-vehModel.maxSocMph)/(vehModel.minSocMph-vehModel.maxSocMph));
					float curSocTarget = vehModel.maxSoc + percToLowSocMph*(
							((vehModel.maxSoc-vehModel.minSoc)*vehModel.percUsableSocPassBuf + 
									vehModel.minSoc) - vehModel.maxSoc);
					
					float mxT2 =0f;
					if (highAccFCOn) {
						mxT2 = (absSoC-(vehModel.minSoc+vehModel.percHighAccBuf))*100f*vehModel.chgKwPerPercOff;
					} else {
						if (absSoC >= curSocTarget) {
							mxT2 = (absSoC-curSocTarget)*100f*vehModel.assistKwPerPercOff;
						}
						else {
							mxT2 = (absSoC-curSocTarget)*100f*vehModel.chgKwPerPercOff;
						}
					}
					float keDesiredDischgKw = Math.min(vehModel.maxMotorKw, 
							Math.max(-vehModel.maxMotorKw,mxT2));
		
					essDesiredDischgKw = keDesiredDischgKw;
					break;
				}
			}
			else essDesiredDischgKw = allElectEssKW;
			

			float essDesiredMcElecKwIn = essDesiredDischgKw - vehModel.auxKw - addAuxPower;
			float mcDesiredKwOut = 0f;
			float zTol = (float)2.0e-5;
			if (essDesiredMcElecKwIn!=0) {
				if (Math.abs(essDesiredMcElecKwIn-mcKwReqFromEss4AE)<=zTol) {
					mcDesiredKwOut = achievedTransPowerInKW;
				}
				else if (essDesiredMcElecKwIn > 0) {
					mcDesiredKwOut = mtModel.outputPowerKW(essDesiredMcElecKwIn);
				} else {
					mcDesiredKwOut = -mtModel.inputPowerKW(-essDesiredMcElecKwIn);
				}
			}
			
			float minMcKw2HelpFc = achievedTransPowerInKW - limFCOut;
			achievedMotorKWOut = Math.min(Math.max(minMcKw2HelpFc, Math.max(mcDesiredKwOut, -curMaxMechMcKwIn)), curMaxMcKwOut);
			if (achievedMotorKWOut > achievedTransPowerInKW) achievedMotorKWOut = achievedTransPowerInKW;
		}
		
		return achievedMotorKWOut;
	}
	
	//Function to return a header string (used when saving array to CSV)
	public static String getHeaderString() {
		return "time,speedMPH,speedMS,roadGrade,distSLSMeters,distSLSMiles,"+
				"dragKW,ascentKW,rollResKW,accelKW,fricBrakeKW,"+
				"fcPowerOut,fcPowerIn,fcEffn,fuelUseSLS,"+
				"mtPowerOut,mtPowerIn,mtEffn,regenKW,batteryUseSLS,relSoC,absSoC,speedTraceSlipMPH";
	}
	//Function to code values into a CSV string
	@Override public String toString() {
		return ""+time+","+speedMPH+","+speedMS+","+roadGrade+","+distSLSMeters+","+distSLSMiles+","+
				dragKW+","+ascentKW+","+rollResKW+","+accelKW+","+fricBrakeKW+","+
				fcPowerOut+","+fcPowerIn+","+fcEffn+","+fuelUseSLS+","+
				mtPowerOut+","+mtPowerIn+","+mtEffn+","+regenKW+","+batteryUseSLS+","+relSoC+","+absSoC+","+speedTraceSlipMPH;
	}
	//Function to save array to file
	public static void writeArrayToFile(String fname, VehicleState[] aRec) {
		String lsep = System.getProperty("line.separator");
		try {
			FileWriter fWriter = new FileWriter(fname);
			fWriter.append(getHeaderString());
			fWriter.append(lsep);
			
			for (int i=0; i<aRec.length; i++) {
				fWriter.append(aRec[i].toString());
				fWriter.append(lsep);
			}
		
			fWriter.flush();
			fWriter.close();
		} catch (IOException e) {}
	}
}
